module Exam.Exercise1.SoundnessSubstitution where

open import Relation.Binary.PropositionalEquality using (_≡_; _≢_; refl; sym)
open import Relation.Nullary                      using (Dec; yes; no; ¬_)
open import Data.String                           using (_≟_)
open import Data.Empty                            using (⊥; ⊥-elim)
open import Function                              using (_∘_)

open import Exam.Exercise1.Syntax
open import Exam.Exercise1.Substitution
open import Exam.Exercise1.Typing

-- Preservation ----------------------------------------------------------------

-- * Intermezzo: rename, weaken, drop, swap

-- A context is a renaming of another context if it can satisfy the same
-- bindings as the other, formally:
--
--   Δ is a renaming of Γ if
--   (∀ x → Γ ∋ x ⦂ T → Δ ∋ x ⦂ T)

ext :
  ∀ { Γ Δ } { x } { A } →
  (∀ {y T} → Γ ∋ y ⦂ T → Δ ∋ y ⦂ T) →
  (∀ {y T} → (Γ , x ⦂ A) ∋ y ⦂ T → (Δ , x ⦂ A) ∋ y ⦂ T)

ext ρ Z = Z
ext ρ (S y≢x y∈Γ) = S y≢x (ρ y∈Γ)

rename :
  ∀ { Γ Δ } →
  (∀ {x T} → Γ ∋ x ⦂ T → Δ ∋ x ⦂ T) →
  ----------------------------------------
  (∀ {M T} → Γ ⊢ M ⦂ T → Δ ⊢ M ⦂ T)

rename ρ (var x⦂T) = var (ρ x⦂T)
rename ρ (⇒-I ⊢M) = ⇒-I (rename (ext ρ) ⊢M)
rename ρ (⇒-E ⊢M ⊢M₁) = ⇒-E (rename ρ ⊢M) (rename ρ ⊢M₁)
rename ρ ℕ-I-zero = ℕ-I-zero
rename ρ (ℕ-I-suc ⊢M) = ℕ-I-suc (rename ρ ⊢M)
rename ρ (ℕ-E-case ⊢L ⊢M ⊢N) = ℕ-E-case (rename ρ ⊢L) (rename ρ ⊢M) (rename (ext ρ) ⊢N)
rename ρ (μ-E ⊢M) = μ-E (rename (ext ρ) ⊢M)

weaken :
  ∀ { Γ } { N } { A } →
  (⊢N : ∅ ⊢ N ⦂ A) →
  Γ ⊢ N ⦂ A

weaken ⊢N = rename (λ{ () }) ⊢N

drop :
  ∀ {Γ} {y} {A B} →
  (∀ {x T} → ((Γ , y ⦂ A) , y ⦂ B) ∋ x ⦂ T → (Γ , y ⦂ B) ∋ x ⦂ T)

drop Z = Z
drop (S x≢y Z) = ⊥-elim (x≢y refl)
drop (S x≢y (S x≢y₁ x∈Γ)) = S x≢y x∈Γ

swap :
  ∀ { Γ } { y z } { A B } →
  (y≢z : z ≢ y) →
  (∀ { x T } → Γ , y ⦂ A , z ⦂ B ∋ x ⦂ T → Γ , z ⦂ B , y ⦂ A ∋ x ⦂ T)

swap y≢z Z = S y≢z Z
swap y≢z (S x≢z Z) = Z
swap y≢z (S x≢z (S x≢y x⦂t)) = S x≢y (S x≢z x⦂t)

subst-preserve :
  ∀ { Γ } { M N } { x } { A T }
  (⊢N      : ∅ ⊢ N ⦂ A) →
  (⊢M      : Γ , x ⦂ A ⊢ M ⦂ T) →
  -------------------------------
  Γ ⊢ M [ x := N ] ⦂ T

subst-preserve {x = x} ⊢N (var{x = y} y⦂T)
  with x ≟ y
subst-preserve {x = x} ⊢N (var {x = _} Z) | no ¬p = ⊥-elim (¬p refl)
subst-preserve {x = x} ⊢N (var {x = _} (S x₁ y⦂T)) | no ¬p = var y⦂T
subst-preserve {x = x} ⊢N (var {x = _} Z) | yes refl = weaken ⊢N
subst-preserve {x = x} ⊢N (var {x = _} (S x≢x y⦂T)) | yes refl = ⊥-elim (x≢x refl)
subst-preserve {x = x} ⊢N (⇒-I{ x = y} ⊢M)
  with x ≟ y
... | yes refl = ⇒-I (rename drop ⊢M)
... | no x≢y = ⇒-I (subst-preserve ⊢N (rename (swap (λ y≡x → x≢y (sym y≡x))) ⊢M))
subst-preserve ⊢N (⇒-E ⊢M ⊢M₁) = ⇒-E (subst-preserve ⊢N ⊢M) (subst-preserve ⊢N ⊢M₁)
subst-preserve ⊢N ℕ-I-zero = ℕ-I-zero
subst-preserve ⊢N (ℕ-I-suc ⊢M) = ℕ-I-suc (subst-preserve ⊢N ⊢M)
subst-preserve {x = x } ⊢N (ℕ-E-case{x = y} ⊢L ⊢M₁ ⊢M₂)
  with x ≟ y
... | yes refl = ℕ-E-case (subst-preserve ⊢N ⊢L) (subst-preserve ⊢N ⊢M₁) (rename drop ⊢M₂)
... | no x≢y  = ℕ-E-case (subst-preserve ⊢N ⊢L) (subst-preserve ⊢N ⊢M₁) (subst-preserve ⊢N (rename (swap (x≢y ∘ sym)) ⊢M₂))
subst-preserve {x = x} ⊢N (μ-E{x = y} ⊢M)
  with x ≟ y
... | yes refl = μ-E (rename drop ⊢M)
... | no x≢y  = μ-E (subst-preserve ⊢N (rename (swap (x≢y ∘ sym)) ⊢M))
