module Exam.Exercise1.Typing where

open import Relation.Binary.PropositionalEquality using (_≡_; _≢_)
open import Exam.Exercise1.Syntax

data Type : Set where
  _⇒_ : Type → Type → Type
  `ℕ  : Type

infixl 5 _,_⦂_

data Context : Set where
  ∅ : Context
  _,_⦂_ : Context → Id → Type → Context

-- * -- the typing judgment
-- this is about defining the relation between Context, Term, and Type
-- written as Γ ⊢ M ⦂ A

variable
  Γ : Context
  A B : Type

-- auxiliary judgment
-- a relation to find a type assumption in the context
-- Γ ∋ x ⦂ A

infix 4 _∋_⦂_

data _∋_⦂_ : Context → Id → Type → Set where

  -- variable appears at top of context
  Z :
    Γ , x ⦂ A ∋ x ⦂ A

  -- variable appears somewhere deeper down
  S :
    x ≢ y →
    Γ ∋ x ⦂ A →
    --------------------
    Γ , y ⦂ B ∋ x ⦂ A

infix 4 _⊢_⦂_

data _⊢_⦂_ : Context → Term → Type → Set where

  var :
    Γ ∋ x ⦂ A →
    -----------
    Γ ⊢ ` x ⦂ A

  -- rules for functions

  ⇒-I :
    Γ , x ⦂ A ⊢ M ⦂ B →
    --------------------
    Γ ⊢ ƛ x ⇒ M ⦂ A ⇒ B

  ⇒-E :
    Γ ⊢ M ⦂ A ⇒ B →
    Γ ⊢ N ⦂ A →
    ----------------
    Γ ⊢ M · N ⦂ B

  -- rules for terms denoting numbers

  ℕ-I-zero :
    Γ ⊢ `zero ⦂ `ℕ

  ℕ-I-suc :
    Γ ⊢ M ⦂ `ℕ →
    --------------------
    Γ ⊢ `suc M ⦂ `ℕ

  ℕ-E-case :
    Γ ⊢ L ⦂ `ℕ →
    Γ ⊢ M ⦂ A →
    Γ , x ⦂ `ℕ ⊢ N ⦂ A →
    ----------------------------------------
    Γ ⊢ case L [zero⇒ M |suc x ⇒ N ] ⦂ A

  μ-E :
    Γ , x ⦂ A ⊢ M ⦂ A →
    --------------------
    Γ ⊢ μ x ⇒ M ⦂ A
