module Exam.Exercise2.SoundnessSubstitution where

open import Relation.Binary.PropositionalEquality using (_≡_; _≢_; refl; sym)
open import Relation.Nullary                      using (Dec; yes; no; ¬_)
open import Data.String                           using (_≟_)
open import Data.Empty                            using (⊥; ⊥-elim)
open import Function                              using (_∘_)

open import Exam.Exercise2.Syntax
open import Exam.Exercise2.Substitution
open import Exam.Exercise2.SemanticsSmallStep     using (Store)
open import Exam.Exercise2.Typing
open import Exam.Exercise2.Sublist

-- Since our extended language, still contains variables and functions,
-- we still need the `subst-preserve` lemma from before.
--
-- Luckily, references are largely independent from variables and functions,
-- so we need precisely the same sub-lemmas as before.
--
-- However, the `rename` and `subst-preserve` lemmas pattern match on terms, so
-- they now have new cases which you need to take care of.

-- * Intermezzo: rename, weaken, drop, swap

ext :
  ∀ { Γ Δ } { x } { A } →
  (∀ {y T} → Γ ∋ y ⦂ T → Δ ∋ y ⦂ T) →
  (∀ {y T} → (Γ , x ⦂ A) ∋ y ⦂ T → (Δ , x ⦂ A) ∋ y ⦂ T)

ext ρ Z = Z
ext ρ (S y≢x y∈Γ) = S y≢x (ρ y∈Γ)

postulate
  rename :
    ∀ { Γ Δ } →
    (∀ {x T} → Γ ∋ x ⦂ T → Δ ∋ x ⦂ T) →
    ----------------------------------------
    (∀ {M T} → Σ ; Γ ⊢ M ⦂ T → Σ ; Δ ⊢ M ⦂ T)

-- Old cases:

-- rename ρ (var x⦂T)           = var (ρ x⦂T)
-- rename ρ (⇒-I ⊢M)            = ⇒-I (rename (ext ρ) ⊢M)
-- rename ρ (⇒-E ⊢M ⊢M₁)        = ⇒-E (rename ρ ⊢M) (rename ρ ⊢M₁)
-- rename ρ ℕ-I-zero            = ℕ-I-zero
-- rename ρ (ℕ-I-suc ⊢M)        = ℕ-I-suc (rename ρ ⊢M)
-- rename ρ (ℕ-E-case ⊢L ⊢M ⊢N) = ℕ-E-case (rename ρ ⊢L) (rename ρ ⊢M) (rename (ext ρ) ⊢N)
-- rename ρ (μ-E ⊢M)            = μ-E (rename (ext ρ) ⊢M)

-- New cases:

-- rename ρ (ref-I-new ⊢M)      = {!!}
-- rename ρ (ref-I-loc ℓ∈Σ)     = {!!}
-- rename ρ (ref-E-get ⊢M)      = {!!}
-- rename ρ (ref-E-set ⊢M₁ ⊢M₂) = {!!}
-- rename ρ ⊤-I                 = {!!}

weaken :
  ∀ { Γ } { N } { A } →
  (⊢N : Σ ; ∅ ⊢ N ⦂ A) →
  Σ ; Γ ⊢ N ⦂ A

weaken ⊢N = rename (λ{ () }) ⊢N

drop :
  ∀ {Γ} {y} {A B} →
  (∀ {x T} → ((Γ , y ⦂ A) , y ⦂ B) ∋ x ⦂ T → (Γ , y ⦂ B) ∋ x ⦂ T)

drop Z = Z
drop (S x≢y Z) = ⊥-elim (x≢y refl)
drop (S x≢y (S x≢y₁ x∈Γ)) = S x≢y x∈Γ

swap :
  ∀ { Γ } { y z } { A B } →
  (y≢z : z ≢ y) →
  (∀ { x T } → Γ , y ⦂ A , z ⦂ B ∋ x ⦂ T → Γ , z ⦂ B , y ⦂ A ∋ x ⦂ T)

swap y≢z Z = S y≢z Z
swap y≢z (S x≢z Z) = Z
swap y≢z (S x≢z (S x≢y x⦂t)) = S x≢y (S x≢z x⦂t)

postulate
  subst-preserve :
    ∀ { Σ } { Γ } { M N } { x } { A T } →
    Σ ; ∅ ⊢ N ⦂ A →
    Σ ; (Γ , x ⦂ A) ⊢ M ⦂ T →
    -------------------------------------
    Σ ; Γ ⊢ M [ x := N ] ⦂ T

-- Old cases:

-- subst-preserve {x = x} ⊢N (var {x = y} y⦂T)       with x ≟ y
-- subst-preserve {x = x} ⊢N (var {x = _} Z)           | no ¬p = ⊥-elim (¬p refl)
-- subst-preserve {x = x} ⊢N (var {x = _} (S x₁ y⦂T))  | no ¬p = var y⦂T
-- subst-preserve {x = x} ⊢N (var {x = _} Z)           | yes refl =  weaken ⊢N
-- subst-preserve {x = x} ⊢N (var {x = _} (S x≢x y⦂T)) | yes refl = ⊥-elim (x≢x refl)
-- subst-preserve {x = x} ⊢N (⇒-I{ x = y} ⊢M)
--   with x ≟ y
-- ... | yes refl = ⇒-I (rename drop ⊢M)
-- ... | no x≢y = ⇒-I (subst-preserve ⊢N (rename (swap (λ y≡x → x≢y (sym y≡x))) ⊢M))
-- subst-preserve ⊢N (⇒-E ⊢M ⊢M₁) = ⇒-E (subst-preserve ⊢N ⊢M) (subst-preserve ⊢N ⊢M₁)
-- subst-preserve ⊢N ℕ-I-zero = ℕ-I-zero
-- subst-preserve ⊢N (ℕ-I-suc ⊢M) = ℕ-I-suc (subst-preserve ⊢N ⊢M)
-- subst-preserve {x = x } ⊢N (ℕ-E-case{x = y} ⊢L ⊢M₁ ⊢M₂)
--   with x ≟ y
-- ... | yes refl = ℕ-E-case (subst-preserve ⊢N ⊢L) (subst-preserve ⊢N ⊢M₁) (rename drop ⊢M₂)
-- ... | no x≢y  = ℕ-E-case (subst-preserve ⊢N ⊢L) (subst-preserve ⊢N ⊢M₁) (subst-preserve ⊢N (rename (swap (x≢y ∘ sym)) ⊢M₂))
-- subst-preserve {x = x} ⊢N (μ-E{x = y} ⊢M)
--   with x ≟ y
-- ... | yes refl = μ-E (rename drop ⊢M)
-- ... | no x≢y  = μ-E (subst-preserve ⊢N (rename (swap (x≢y ∘ sym)) ⊢M))

-- New cases:

-- subst-preserve {x = x} ⊢N (ref-I-new ⊢M) = {!!}
-- subst-preserve {x = x} ⊢N (ref-E-get ⊢M) = {!!}
-- subst-preserve {x = x} ⊢N (ref-E-set ⊢M₁ ⊢M₂) = {!!}
-- subst-preserve {x = x} ⊢N (ref-I-loc l∈Σ) = {!!}
-- subst-preserve {x = x} ⊢N ⊤-I = {!!}
